/*

Interpreter.java 
  -- classes implementing the core constraint solving operations
     as part of the "ia_solver library" version 0.1beta1, 10/97

Copyright (C) 1997 Timothy J. Hickey

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Library General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

You should have received a copy of the GNU Library General Public
License along with this library; if not, write to the
Free Software Foundation, Inc., 59 Temple Place - Suite 330,
Boston, MA  02111-1307, USA.

 */


package ia_solver;

import java.awt.*;
import java.io.DataInputStream;
import java.lang.*;
import java.util.*;

import ia_parser.*;
import ia_math.RealInterval;
import ia_math.IAMath;
/* import ia_solver.utility.*;*/


/**
 * this class implements the core constraint solving
 * operations. It is meant to be called from a user interface
 * which could be an applet, GUI application, or a 
 * text-based application. The user interface is abstracted
 * as an "interface" class which define methods that allow
 * the interpreter to have access to the constraint set,
 * variable list, etc.
 */
public class Interpreter extends Thread
{
  public static final int NARROW = 0;
  public static final int ABSOLVE = 1;

  //  private Interface parent;
  private SolverGUI gui;
  private String program="";
  private int action = NARROW;

  Interpreter(String s, SolverGUI gui)
  {
    program = s;
    this.gui = gui;
  }

  Interpreter(SolverGUI gui)
  {
    this.gui = gui;
  }

  Interpreter(int action, SolverGUI gui)
  {
    this.action = action;
    this.gui = gui;
  }

  public void run(){
    if (action == NARROW) {
      gui.setMessageField("narrowing ....");
      narrow();
    }
    else if (action == ABSOLVE) {
      gui.setMessageField("absolving ....");
      absolve();
    }
    else
      gui.setMessageField(
       "ERROR in Interpreter ... unknown action parameter");
  }


  /**
   * This returns true if the tree can be narrowed max_narrow times.
   */
  public boolean narrow(Exp tree,int max_narrow)
  {
    if (tree==null) return false;

    for (int i=0; i<=max_narrow; i++) {
      gui.setMessageField("narrowing iteration # "+ i);
      if (!tree.narrow()) {
        return false;
      }
      gui.notifyNarrows(i);
    }

    return true;
  }

  public boolean narrow(Exp tree,int max_narrow,boolean show)
  {
    if (tree==null) return false;

    for (int i=0; i<=max_narrow; i++) {
      if (show) gui.setMessageField("narrowing iteration # "+ i);
      if (!tree.narrow()) {
        return false;
      }
      if (show) gui.notifyNarrows(i);
    }

    return true;
  }


  /**
   * This attempts to narrow each variable using absolve_var.
   */
  void absolve() {
    RealIntervalTable vars = gui.getVars();
    Exp tree = gui.getConstraints();
    String[] var_names = vars.getNames();

    if (tree==null) {
      gui.setMessageField("Syntax Error.....");
      return;
    }

    gui.setButtonStatus("ABSOLVE");

    for(int i=0; i<var_names.length; i++) {
      gui.setMessageField("absolve iteration # "+i);
      if (absolve_var(var_names[i],vars))
          gui.notifyNarrows(0);
    }
    gui.setMessageField("absolve completed");
    gui.setButtonStatus("READY");

  }

  boolean absolve_var(String v, RealIntervalTable vars) {
    Exp tree = gui.getConstraints();
    if (tree==null) {
      gui.setMessageField("Syntax error.....");
      return false;
    }

    boolean absolved_hi=false,absolved_lo=false;;
    
    absolved_hi = absolve_var_hi(v,vars,tree,0.5);
    if (absolved_hi) {
      absolved_lo = absolve_var_lo(v,vars,tree,0.5);
      if (!(absolved_lo))
        absolved_lo = absolve_var_lo(v,vars,tree,0.06125);
    }
    else {
      absolved_hi = absolve_var_hi(v,vars,tree,0.93875);
      absolved_lo=absolve_var_lo(v,vars,tree,0.5);
      if (!(absolved_lo))
        absolved_lo = absolve_var_lo(v,vars,tree,0.06125);
    }
    return (absolved_hi || absolved_lo);
  }


  void splitVarHi(RealInterval var, double fraction) {
    double splitPoint = var.lo()*fraction+ var.hi()*(1.0-fraction);
    if (Double.isInfinite(var.lo()) || Double.isInfinite(var.hi()))
      splitPoint = IAMath.midpoint(var).hi();
    var.intersect(new RealInterval(Math.min(var.hi(),splitPoint),
                                   var.hi()));
  }

  void splitVarLo(RealInterval var, double fraction) {
    double splitPoint = var.lo()*fraction+ var.hi()*(1.0-fraction);
    if (Double.isInfinite(var.lo()) || Double.isInfinite(var.hi()))
      splitPoint = IAMath.midpoint(var).lo();
    var.intersect(new RealInterval(var.lo(),
                                   Math.max(var.lo(),splitPoint)));
  }


  boolean absolve_var_hi(String v,RealIntervalTable vars, Exp tree, double fraction) {
    RealIntervalTable myVars = (RealIntervalTable) vars.clone();
    RealInterval var = myVars.lookup(v);
    int max_narrow = (gui.getMaxNarrow()).intValue();

    gui.setMessageField("absolve hi: var= "+v+" fraction = "+fraction);

    splitVarHi(var,fraction);
    tree.bindVars(myVars);
    
    boolean failed = !narrow(tree,max_narrow,false);
    if (failed) { // in this case we can narrow the variable down
      var = vars.lookup(v);
      splitVarLo(var,fraction);
    }       
    tree.bindVars(vars);
    return failed;
  }

  boolean absolve_var_lo(String v,RealIntervalTable vars, Exp tree, double fraction) {
    RealIntervalTable myVars = (RealIntervalTable) vars.clone();
    RealInterval var = myVars.lookup(v);
    int max_narrow = (gui.getMaxNarrow()).intValue();

    gui.setMessageField("absolve lo: var= "+v+" fraction = "+fraction);

    splitVarLo(var,fraction);
    tree.bindVars(myVars);
    boolean failed = !narrow(tree,max_narrow,false);
    if (failed) { // in this case we can narrow the variable up
      var = vars.lookup(v);
      splitVarHi(var,fraction);
    }       
    tree.bindVars(vars);
    return failed;
  }

  void narrow()
  {
    Exp tree;
    RealInterval result;
    boolean narrow_result;
    int max_narrow=100;
    int i,j;

    gui.setButtonStatus("NARROW");

    tree = gui.getConstraints();
    if (tree==null) {
        gui.notifyFailure();
        gui.setMessageField("Syntax error......");
        gui.setButtonStatus("READY");
        return;
    }

    max_narrow = (gui.getMaxNarrow()).intValue();

    if (narrow(tree,max_narrow)){
        gui.notifySuccess();
	gui.setMessageField("narrowing was successful");
    }
    else {
        gui.notifyFailure();
        gui.setMessageField("narrowing failed!");
    }

        gui.setButtonStatus("READY");
  }





};




